/* Usart adapter compatible with 6502 bus
*  It has fixed baudrate 57600 BPS
* clk: input clock
* rst: negative reset signal
* tx_pin: transmission pin USART
* rx_pin: receive pin USART
* addr: address selector
* cs: chip select
* we: write enable
* di: data input bus
* do: data output bus
* addr: input address
* 	00 - DR (Data register, WR,RD)
* 	01 - SR (Status register, RO)
* 	02 - CR (Control register, WR) 
*/

module tiny_usart(
	input clk,
	input rst,
	output tx_pin,
	input rx_pin,
	input [2:0] addr,
	input cs,
	input we,
	input [7:0] di,
	output [7:0] do
);
	reg utx_valid;
	wire utx_ready;
	reg [7:0] tx_dr;
	reg [7:0] rx_dr_rdy;
	reg [7:0] idata;
	wire [7:0] urx_bus;
	wire urx_valid;
	reg urx_ready;
	wire urx_overflow;

	localparam DR_ADDR = 3'd0;
	localparam SR_ADDR = 3'd1;
	localparam CR_ADDR = 3'd2;

	always @(posedge clk or negedge rst)
		if(~rst) begin
			tx_dr <= 8'd0;
			utx_valid <= 1'b0;
			urx_ready <= 1'b0;
			rx_dr_rdy <= 1'b0;
		end 
		else begin
			if(cs && we) begin
				if(addr==DR_ADDR) begin
					tx_dr <= di;
				end
				else if(addr==CR_ADDR) begin
					utx_valid <= di[0];
				end
			end
			else if(cs && !we) begin
				if(addr==SR_ADDR) begin
					idata <= { 5'd0, urx_overflow, rx_dr_rdy, utx_ready };
				end
				else if(addr==CR_ADDR) begin
					idata <= { 7'd0, utx_valid };
				end
				else if(addr==DR_ADDR) begin
					idata <= urx_bus;
					rx_dr_rdy <= 1'b0;
				end
			end
			if(urx_valid) begin
				urx_ready <= 1'b1;
				rx_dr_rdy <= 1'b1;
			end

			if(utx_valid) utx_valid <= 1'b0;
			if(urx_ready) urx_ready <= 1'b0;
		end


	// Transmit component
	usart_tx_fifo #(.SPEED_NDIV(13)) utx( 
		.clk(clk), .rst(rst),
		.valid(utx_valid), .ready(utx_ready),
		.data(tx_dr), .u_tx_pin(tx_pin)
	);

	// Receive component
	usart_rx_fifo #(.SPEED_NDIV(13)) urx( .clk(clk), .rst(rst),
		.valid(urx_valid), .ready(urx_ready),
		.overflow(urx_overflow), .data(urx_bus),
		.u_rx_pin(rx_pin)
	);
   
	reg cs_del;
	always @(posedge clk or negedge rst)
		if(~rst) cs_del <= 1'b0;
		else cs_del <= cs;

	tristate_buf tbuf( 
		.oe(cs_del),
		.in(idata),
		.out(do)
	);
	
endmodule
