/* Generic bidir fifo */

module synchronous_fifo
	#( parameter DEPTH=4, DWIDTH=8 )
	( input clk, input rst, 			 /* Reset and clock */
      input wr_en, input rd_en,	 		 /* Read and write signals */
	  input [DWIDTH-1:0] data_in, 
	  output reg [DWIDTH-1:0] data_out,  /* Input and output data */
	  output empty, output full 		 /* Empty and full signals */
	);
	
	// Local parameter
	localparam DW = (DWIDTH  < 1) ? 1 : DWIDTH;
	localparam AW = (DEPTH < 1) ? 1 : DEPTH;
	localparam MS = 2**DEPTH;

	reg [AW:0] write_pointer;
	reg [AW:0] read_pointer;
	reg [DW-1:0] memory[0:MS];

	wire empty_int = (write_pointer[AW] == read_pointer[AW]);
	wire full_or_empty = (write_pointer[AW-1:0] == read_pointer[AW-1:0]);
	
	assign full  = full_or_empty & !empty_int;
	assign empty = full_or_empty & empty_int;
	
	wire [AW-1:0] addr_w = write_pointer[AW-1:0];
	wire [AW-1:0] addr_r = read_pointer[AW-1:0];
	
	
	// Read counter update
	always @(posedge clk or negedge rst) begin
		if (~rst) begin
			read_pointer  <= 0;
		end
		else if (rd_en) begin
			read_pointer <= read_pointer + 1'd1;
		end
	end

	// Write counter update
	always @(posedge clk or negedge rst) begin
		if (~rst) begin
			write_pointer <= 0;
		end
		else if (wr_en) begin
			write_pointer <= write_pointer + 1;
		end
	end

	// Memory (it will be automatic assigned to block ram)
	always @(posedge clk) begin
		if(wr_en)
			memory[addr_w] <= data_in;
		if(rd_en)
			data_out <= memory[addr_r];
	end
	

endmodule

