/* RETRO Microcontroller top module
 * for ice core board. SDRAM controller will be run 
 * without the cache memory in the feature we can improve that
 */

module retro_65c02_mcu (
	input clkin,		// Clock input signal
	output usart_tx, 	// Usart TX and RX
	input usart_rx,
	output [3:0] led // LED interface
);

	wire clk;				// 3.25MHz clock
	wire rst;				// Reset negative generator


	// Clock generator block
	clkgen clk_blk( 
		.clock_in(clkin),
		.clk(clk),
		.rst(rst)
	);

	//6502 CPU
	wire cpu_nmi = 1'b1;
	wire cpu_irq = 1'b1;
	wire cpu_rdy = 1'b1;
	wire cpu_we;
	wire [7:0] cpu_di;
	wire [7:0] cpu_do;
	wire [15:0] cpu_addr;
	
	cpu_65c02 cpu_core(
		.clk(clk),
		.reset(~rst),
		.AB(cpu_addr),
		.DI(cpu_di),
		.DO(cpu_do),
		.WE(cpu_we), // 0 - read, 1 - write
		.IRQ(cpu_irq),
		.NMI(cpu_nmi),
		.RDY(cpu_rdy)
	);

	// Main address decoder
	wire ram_cs; wire rom_cs; // Decoder chip sel
	wire periph_cs;	//Decoder chip sel

	main_address_decoder addr_decode(
		.addr_bus(cpu_addr),
		.rom_cs(rom_cs),
		.ram_cs(ram_cs),
		.periph_cs(periph_cs)
	);
	// Internal ROM memory
	internal_rom #(.BITS(13)) rom_mem (
		.clk(clk),
		.rst(rst),
		.oe(rom_cs),
		.addr(cpu_addr[12:0]),
		.data(cpu_di)
	);
	// Internal RAM memory
	internal_ram #(.BITS(12)) ram_mem(
		.clk(clk),
		.rst(rst),
		.we(cpu_we),
		.oe(ram_cs),
		.addr(cpu_addr[11:0]),
		.din(cpu_do),
		.dout(cpu_di)
	);
	


	//Peripheral address decoder
	wire gpio_pcs, uart_pcs;
	device_address_decoder dev_decoder(
		.addr_in(cpu_addr[7:0]),
		.periph_cs(periph_cs),
		.periph1(gpio_pcs),
		.periph2(uart_pcs)
	);
	
	// Module for output LED diode
	output_gpio gpio_ctrl(
		.clk(clk),
		.rst(rst),
		.di(cpu_do),
		.oe(gpio_pcs),
		.we(cpu_we),
		.port(led)
	);
	
	//Simple UART
	tiny_usart usart(
		.clk(clk),
		.rst(rst),
		.tx_pin(usart_tx),
		.rx_pin(usart_rx),
		.addr(cpu_addr[2:0]),
		.cs(uart_pcs),
		.we(cpu_we),
		.di(cpu_do),
		.do(cpu_di)
	);
	
endmodule
