/* Simple verilog USART module 
*  Output frequency should be Fout = (Fin * K)/(2**N)
*
*/

/** Usart Baudrate prescaller 
* Parameters:
* 	K - multiplier
*   N - divider
* Wires:
*  clk_in - Input clock
*  clk_rx - clock for receiver
*  clk_tx - clock for transmitter
*/
module usart_brr #(parameter K=1,N=1)
	( input clk_in, input rst, output clk_out );
	reg [N-1:0] phase_acc; 	//Phase accumulator
	// Master DDS
	always @(posedge clk_in or negedge rst) begin
		if(~rst) phase_acc <= {N{1'b0}};
		else phase_acc <= phase_acc + K;
	end
	assign clk_out = phase_acc[N-1];
endmodule


module usart_tx_div(input inclk, input rst, output outclk);
	reg [3:0] div_16;		//Phase divider
	assign outclk = div_16[3];
	always @(posedge inclk or negedge rst)
		if(~rst) div_16 <= 4'b0;
		else div_16 <= div_16 + 1;
endmodule

/* USART transmit module
*	clk  - Transmitter clock
*	rst  - Reset signal
*	en   - Start transmission
*	tx   - Output usart
*	rdy  - Transmit ack
*/
module usart_tx( 
	input clk,
	input rst,
	input en,
	input [7:0] data,
	output reg tx,
	output reg rdy 
);
	
	reg [3:0] bit_cnt; 	// Bit counter
	reg [1:0] en_sync;	// Enable sync register
	reg [7:0] tx_sreg;	// Transmit shift register

	// Inter domain clock synchronizer
	always @(posedge clk) begin
		en_sync[0] <= en;
		en_sync[1] <= en_sync[0];
	end

	wire d_en = en_sync[1];
	
	// Usart main process
	always @(posedge clk or negedge rst) begin
		if( ~rst ) 
		begin	// When reset on uart
			rdy <= 1'b1;
			bit_cnt <= 4'b0;
			tx <= 1'b1;
		end 
		else 
		begin 	// When normal process
			if( d_en & rdy ) 
			begin	// On transmit start
				bit_cnt <= 4'b0;
				rdy <= 1'b0;
				tx <= 1'b0;
				tx_sreg <= data;
			end
			else if( ~rdy )
			begin	// When data is transmited
				if( bit_cnt<8 ) 
				begin
					tx <= tx_sreg[0];
					tx_sreg <= { tx_sreg[0],tx_sreg[7:1] };
				end
				else if( bit_cnt==8 )
					tx <= 1'b1;
				else 
					rdy <= 1'b1;
				bit_cnt <= bit_cnt + 1;
			end
		end	
	end
endmodule



/* USART receive module 
 * clk_in - Input clock 
 * rst - reset signal. (Active low)
 * ack_in - read acknowledge 
 * rxd_in -  input RX data
 * data - output data
 * rdy - New data in buffer arrived
*/

module usart_rx( input clk,
	input rst, input ack_in, input rxd_in,
	output [7:0] data, output reg rdy );

	localparam STATE_IDLE  		= 2'd0;
	localparam STATE_START 		= 2'd1;
	localparam STATE_RECEIVE    = 2'd2;
	localparam STATE_STOP 		= 2'd3;

	/* Helper function for convert data */
	function automatic to_bit(input [2:0] inb);
		(* full_case,parallel_case *)
		case (inb)
			3'b000: to_bit = 1'b0;
			3'b001: to_bit = 1'b0;
			3'b010: to_bit = 1'b0;
			3'b011: to_bit = 1'b1;
			3'b100: to_bit = 1'b0;
			3'b101: to_bit = 1'b1;
			3'b110: to_bit = 1'b1;
			3'b111: to_bit = 1'b1;
		endcase
	endfunction


	reg [1:0] rxd_sync, ack_sync;
	reg [7:0] dreg;

	assign data = dreg[7:0];

	// Inter domain clock synchronizer
	always @(posedge clk) begin
		rxd_sync[0] <= rxd_in;
		rxd_sync[1] <= rxd_sync[0];
		pedge <= rxd_sync[1];
		ack_sync[0] <= ack_in;
		ack_sync[1] <= ack_sync[0];
	end

	wire ack = ack_sync[1];
	wire rxd = rxd_sync[1];

	reg [1:0] state;
	reg [3:0] scnt; 
	reg [2:0] bcnt;
	reg [2:0] bits;
	reg pedge;
	//Main process for receiver
	always @(posedge clk or negedge rst)
		if(~rst) begin
			scnt <= 4'b0;
			bcnt <= 3'b0;
			rdy <= 1'b0;
			state <= STATE_IDLE;
		end
		else begin /* RST */
			if(state==STATE_IDLE & ack) rdy <= 1'b0;
			(* full_case,parallel_case *)
			case(state)
				STATE_IDLE: if(~rxd & pedge) begin
					state <= STATE_START;
					scnt <= 4'b0;
					bcnt <= 3'b0;
					rdy <= 1'b0;
				end /* STATE_IDLE */
				STATE_START: begin
					if(scnt < 4'd14) scnt <= scnt + 1;
					else begin 
						scnt <= 4'b0;
						state <= STATE_RECEIVE;
					end
				end /* STATE_START */
				STATE_RECEIVE:  begin
					if(scnt < 4'd15) begin
						if(scnt==4'd4)  bits[2] <= rxd;
						if(scnt==4'd8)  bits[1] <= rxd;
						if(scnt==4'd11) bits[0] <= rxd;
						scnt <= scnt + 1;
					end 
					else begin	/*scnt*/
						scnt <= 4'd0;
						dreg <= {to_bit(bits),dreg[7:1]};
						if(bcnt<3'd7) bcnt <= bcnt + 1;
						else begin
							state <= STATE_STOP;
							scnt  <= 4'b0;
						end /* bcnt */
					end /*scnt*/
				end /* STATE RECEIVE */
			    STATE_STOP: begin
					if(scnt <= 4'd6) scnt <= scnt + 1;
					else begin 
						rdy <= 1'b1;
						state <= STATE_IDLE;
					end /* !scnt */
				end /* STATE_STOP*/
			endcase
		end /* RST */
endmodule


/** Usart transmit BUFFER */
module usart_tx_fifo
	#(parameter DEPTH_POW=4, SPEED_KDIV=2416, SPEED_NDIV=15)
	( /* Bus input signals */
	  input clk, input rst, 		/* Clock and reset signals */
	  input valid, output ready, 	/* Data valid and data ready signals */
	  input [7:0] data,				/* Input data signals */
	  output u_tx_pin			    /* Output usart TX pin */
	);

	wire clk_rx;			// Clock for RX
	wire clk_tx;			// Clock for TX
	wire u_tx_rdy;			// Transmit completed
	reg  u_start_xmit;		// Start transmit
	reg f_rd;		   	// Fifo read
	wire [7:0] u_db;		// Usart data bus
	wire f_empty, f_full;	// Fifo empty fifo full	
	// BRR divider
	usart_brr #(.N(SPEED_NDIV),.K(SPEED_KDIV)) usart_presc( 
		.clk_in(clk), .clk_out(clk_rx), .rst(rst)
	);
	// Div16 tx
	usart_tx_div usart_tx_div(
		.inclk(clk_rx), .rst(rst), .outclk(clk_tx) 
    );
	
	// Fifo queue
	synchronous_fifo #(.DEPTH(DEPTH_POW)) tx_fifo(
		.clk(clk), .rst(rst), 
		.wr_en(valid), .rd_en(f_rd),
		.data_in(data), .data_out(u_db),
		.empty(f_empty), .full(f_full)
	);

	// Usart transmit module
	usart_tx u_txm( .clk(clk_tx), .rst(rst),
		.en(u_start_xmit), .data(u_db),
		.tx(u_tx_pin), .rdy(u_tx_rdy)
	);
	assign ready = ~f_full & ~f_rd & ~valid;

	// State machine for control USART transmitter
	reg [1:0] tx_state;
	localparam STATE_WAIT_FOR_FIFO_NE = 2'd0;
	localparam STATE_FIFO_RD_PULSE = 2'd1;
	localparam STATE_UART_WR_ACK = 2'd2;
	localparam STATE_UART_TX_WAIT = 2'd3;
	

	always @(posedge clk or negedge rst)
	if(~rst) begin
		tx_state <= STATE_WAIT_FOR_FIFO_NE;
		f_rd <= 1'b0;
	end
	else begin
		(* full_case,parallel_case *)
		case(tx_state)
		STATE_WAIT_FOR_FIFO_NE: begin
			if(!f_empty && u_tx_rdy) begin
				f_rd <= 1'b1;
				tx_state <= STATE_FIFO_RD_PULSE;
			end
		end
		STATE_FIFO_RD_PULSE: begin
			f_rd <= 1'b0;
			u_start_xmit <= 1'b1;
			tx_state <= STATE_UART_WR_ACK;
		end
		STATE_UART_WR_ACK: begin
			if( ~u_tx_rdy ) begin 
				tx_state <= STATE_UART_TX_WAIT;
				u_start_xmit <= 1'b0;
			end
		end
		STATE_UART_TX_WAIT: begin
			if( u_tx_rdy ) tx_state <= STATE_WAIT_FOR_FIFO_NE;
		end
		endcase
	end /* rst */

endmodule

/* Usart fifo RX module
* addr line = 0 (read data register)
* addr line = 1 (read status register)
* status register:
	* bit0 - data avail
	* bit1 - tx overflow
	* Read data clear the overflow */

module usart_rx_fifo
	#(parameter DEPTH_POW=4, SPEED_KDIV=2416, SPEED_NDIV=15)
	( /* Bus input signals */
	  input clk, input rst, 		/* Clock and reset signals */
	  output valid, input ready, 	/* Data valid and data ready signals */
	  output overflow,				/* Fifo buffer overflow shit happens */
	  output [7:0] data,		    /* Output data signals */
	  input u_rx_pin			    /* Output usart TX pin */
	);

	wire u_rdy; 		// Usart ready
	wire [7:0] u_db8;	// Usart data bus
	wire f_empty;		// Fifo empty
	wire clk_rx;		// Receive clock
	wire u_ack;
	reg fifo_wr;
	reg urdy_pstate;

	assign u_ack = ~overflow;
	assign valid = ~f_empty;

	// BRR divider
	usart_brr #(.N(SPEED_NDIV),.K(SPEED_KDIV)) usart_presc( 
		.clk_in(clk), .clk_out(clk_rx), .rst(rst)
	);

	// Fifo queue
	synchronous_fifo #(.DEPTH(DEPTH_POW)) tx_fifo(
		.clk(clk), .rst(rst), 
		.wr_en(fifo_wr), .rd_en(ready),
		.data_in(u_db8), .data_out(data),
		.empty(f_empty), .full(overflow)
	);

	// Usart receiver
	usart_rx urx( .clk(clk_rx), .rst(rst),
		.ack_in(u_ack), .rxd_in(u_rx_pin),
		.rdy(u_rdy), .data(u_db8)
	);

	always @(posedge clk or negedge rst)
		if(~rst) begin
			fifo_wr <= 1'b0;
			urdy_pstate <= 1'b0;
		end
		else begin
			fifo_wr <= (u_rdy & ~urdy_pstate);
			urdy_pstate <= u_rdy;
		end
endmodule
